"""Module for creation of Project objects"""

from __future__ import annotations
from typing import TYPE_CHECKING
from attr import dataclass

from .project import ProjectMetadata
from .projects.skale_manager import SkaleManagerProject


if TYPE_CHECKING:
    from .project import Project
    from .network import Network


@dataclass
class Projects:
    """Contains all known projects"""
    skale_manager = ProjectMetadata(name='skale-manager', path='skale-manager')


def create_project(network: Network, name: str) -> Project:
    """Create Project object based on it's name"""
    if name == Projects.skale_manager.name:
        return SkaleManagerProject(network, Projects.skale_manager)
    raise ValueError(f'Project with name {name} is unknown')



"""Contains Project class"""
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from attr import dataclass
from eth_utils.address import to_canonical_address
import requests

from .constants import REPOSITORY_URL, NETWORK_TIMEOUT
from .instance import Instance, InstanceData
from .network import ListedNetwork

if TYPE_CHECKING:
    from eth_typing import Address
    from .network import Network


@dataclass
class ProjectMetadata:
    """Contains project metadata"""
    name: str
    path: str


class Project(ABC):
    """Represents set of smart contracts known as project"""

    def __init__(self, network: Network, metadata: ProjectMetadata) -> None:
        super().__init__()
        self.network = network
        self._metadata = metadata

    @abstractmethod
    @property
    def github_repo(self) -> str:
        """URL of github repo with the project"""

    def get_instance(self, alias_or_address: str) -> Instance:
        """Create instance object based on alias or address"""
        if self.network.web3.is_address(alias_or_address):
            address = to_canonical_address(alias_or_address)
            return self.create_instance(address)
        alias = alias_or_address
        url = self.get_instance_data_url(alias)
        response = requests.get(url, timeout=NETWORK_TIMEOUT)
        if response.status_code == 200:
            data = InstanceData.from_json(response.text)
            if len(data.data.values()) != 1:
                raise ValueError(f'Error during parsing data for {alias}')
            address = to_canonical_address(list(data.data.values())[0])
            return self.create_instance(address)
        raise ValueError(f"Can't download data for instance {alias}")

    def download_abi_file(self, version: str) -> str:
        """Download file with ABI"""
        response = requests.get(self.get_abi_url(version), timeout=NETWORK_TIMEOUT)
        return response.text

    def get_abi_url(self, version: str) -> str:
        """Calculate URL of ABI file"""
        return f'{self.github_repo}releases/download/{version}/{self.get_abi_filename(version)}'

    @abstractmethod
    def get_abi_filename(self, version: str) -> str:
        """Return name of a file with ABI"""

    def get_instance_data_url(self, alias: str) -> str:
        """Get URL of a file containing address for provided alias"""
        if isinstance(self.network, ListedNetwork):
            return f'{REPOSITORY_URL}{self.network.path}/{self._metadata.path}/{alias}.json'
        raise ValueError('Network is unknown')

    @abstractmethod
    def create_instance(self, address: Address) -> Instance:
        """Create instance object based on known address"""



"""Module for work with SKALE contracts metadata"""

from __future__ import annotations
from dataclasses import dataclass
import json
from typing import Optional
import requests

from .constants import REPOSITORY_URL, METADATA_FILENAME, NETWORK_TIMEOUT


@dataclass
class NetworkMetadata:
    """Contains metadata of a network"""
    name: str
    chain_id: int
    path: str

@dataclass
class MetadataFile:
    """Represents file with metadata"""
    networks: list[NetworkMetadata]

    @classmethod
    def from_json(cls, data: str) -> MetadataFile:
        """Create MetadataFile object from json string"""
        file = json.loads(data)
        networks = []
        for network in file['networks']:
            networks.append(NetworkMetadata(
                name=network['name'],
                chain_id=network['chainId'],
                path=network['path']))
        return cls(networks)

class Metadata:
    """Class to manage SKALE contracts metadata"""
    networks: list[NetworkMetadata]

    def __init__(self) -> None:
        self.download()

    def download(self) -> None:
        """Download metadata"""
        metadata_response = requests.get(
            REPOSITORY_URL + METADATA_FILENAME,
            timeout=NETWORK_TIMEOUT
        )
        metadata = MetadataFile.from_json(metadata_response.text)
        self.networks = metadata.networks

    def get_network_by_chain_id(self, chain_id: int) -> Optional[NetworkMetadata]:
        """Get network metadata by it's chain id.
        Returns None if there is no such network in the metadata.
        """
        for network in self.networks:
            if network.chain_id == chain_id:
                return network
        return None



"""Module for work with networks"""

from __future__ import annotations
from typing import TYPE_CHECKING
from web3 import Web3
from web3.providers.base import BaseProvider

from .project_factory import create_project


if TYPE_CHECKING:
    from .project import Project
    from .skale_contracts import SkaleContracts


class Network:
    """Represents blockchain with deployed smart contracts projects"""
    def __init__(self, skale_contracts: SkaleContracts, provider: BaseProvider):
        self.web3 = Web3(provider)
        self._skale_contracts = skale_contracts

    def get_project(self, name: str) -> Project:
        """Get Project object by it's name"""
        return create_project(self, name)

    @property
    def skale_contracts(self) -> SkaleContracts:
        """Get SkaleContracts object associated with the network"""
        return self._skale_contracts


class ListedNetwork(Network):
    """Network that is listed in the metadata"""
    def __init__(self, skale_contracts: SkaleContracts, provider: BaseProvider, path: str):
        super().__init__(skale_contracts, provider)
        self.path = path



"""Module for instance management"""

from __future__ import annotations
from abc import ABC, abstractmethod
import json
from typing import TYPE_CHECKING, Optional
from attr import dataclass


if TYPE_CHECKING:
    from eth_typing import Address
    from web3 import Web3
    from web3.contract.contract import Contract
    from .abi import SkaleAbi
    from .project import Project



@dataclass
class InstanceData:
    """Contains instance data"""
    data: dict[str, str]
    @classmethod
    def from_json(cls, data: str) -> InstanceData:
        """Create InstanceData object from json string"""
        return cls(data=json.loads(data))


class Instance(ABC):
    """Represents deployed instance of a smart contracts project"""
    def __init__(self, project: Project, address: Address) -> None:
        self._project = project
        self._version: Optional[str] = None
        self._abi: Optional[SkaleAbi] = None
        self.address = address

    @property
    def web3(self) -> Web3:
        """Get web3 object"""
        return self._project.network.web3

    @property
    def version(self) -> str:
        """Get version of the project instance"""
        if self._version is None:
            self._version = self._get_version()
            if not '-' in self._version:
                self._version = self._version + '-stable.0'
        return self._version

    @property
    def abi(self) -> SkaleAbi:
        """Get abi file of the project instance"""
        if self._abi is None:
            self._abi = json.loads(self._project.download_abi_file(self.version))
        return self._abi

    @abstractmethod
    def get_contract_address(self, name: str) -> Address:
        """Get address of the contract by it's name"""

    def get_contract(self, name: str) -> Contract:
        """Get Contract object of the contract by it's name"""
        address = self.get_contract_address(name)
        return self.web3.eth.contract(address=address, abi=self.abi[name])

    # protected

    @abstractmethod
    def _get_version(self) -> str:
        pass



"""Contains SkaleContracts class"""

from web3 import Web3
from web3.providers.base import BaseProvider

from .metadata import Metadata
from .network import Network, ListedNetwork


class SkaleContracts:
    """Entry point of the SKALE Contracts library"""
    def __init__(self) -> None:
        self.metadata = Metadata()

    def get_network_by_provider(self, provider: BaseProvider) -> Network:
        """Get network by provider"""
        web3 = Web3(provider)
        chain_id = web3.eth.chain_id
        network_metadata = self.metadata.get_network_by_chain_id(chain_id)
        if network_metadata:
            return ListedNetwork(self, provider, network_metadata.path)
        return Network(self, provider)

    def get_network_by_http_endpoint(self, http_endpoint: str) -> Network:
        """Get network by HTTP URI"""
        return self.get_network_by_provider(Web3.HTTPProvider(http_endpoint))
